@@ -2,27 +2,29 @@
 
 It enables papermill to run notebooks against Noteable as though it were executing a notebook locally.
 """
-import asyncio
 import functools
 import json
 import logging
 from contextlib import asynccontextmanager
 from typing import Generator, Optional
 
+import httpx
 import nbformat
 import orjson
 from jupyter_client.utils import run_sync
 from nbclient.exceptions import CellExecutionError
 from nbformat import NotebookNode
 from origami.client import NoteableClient, SkipCallback
-from origami.types.files import NotebookFile
-from origami.types.jobs import (
+from origami.defs.files import NotebookFile
+from origami.defs.jobs import (
     CustomerJobDefinitionReferenceInput,
     CustomerJobInstanceReferenceInput,
     JobInstanceAttempt,
+    JobInstanceAttemptRequest,
     JobInstanceAttemptStatus,
+    JobInstanceAttemptUpdate,
 )
-from origami.types.rtu import (
+from origami.defs.rtu import (
     AppendOutputEventSchema,
     BulkCellStateMessage,
     DisplayHandlerUpdateEventSchema,
@@ -106,6 +108,7 @@ class NoteableEngine(Engine):
         # Currently, this is only used when the output is of type display_data.
         self.__noteable_output_id_cache = {}
         self.file = None
+        self.job_instance_attempt: Optional[JobInstanceAttempt] = None
 
     def catch_cell_metadata_updates(self, func):
         """A decorator for catching cell metadata updates related to papermill
@@ -136,7 +139,7 @@ class NoteableEngine(Engine):
         if not kwargs.get("logger"):
             kwargs["logger"] = logger
 
-        dagster_logger = kwargs["logger"]
+        ext_logger = kwargs["logger"]
 
         dagster_context = kwargs.get("dagster_context")
 
@@ -152,7 +155,13 @@ class NoteableEngine(Engine):
             else:
                 raise ValueError("No file_id or derivable file_id found")
 
-        job_instance_attempt = kwargs.get("job_instance_attempt")
+        job_instance_attempt_request = (
+            JobInstanceAttemptRequest.parse_obj(kwargs.get('job_instance_attempt'))
+            if kwargs.get('job_instance_attempt')
+            else None
+        )
+
+        # Setup job instance attempt from the provided customer job metadata
         if job_metadata := kwargs.get("job_metadata", {}):
             version = await self.client.get_version_or_none(original_notebook_id)
             if version is not None:
@@ -161,7 +170,7 @@ class NoteableEngine(Engine):
                 file = await self.client.get_notebook(original_notebook_id)
                 space_id = file.space_id
 
-            # 1: Ensure the job definition&instance references exists
+            # 1: Ensure the job definition and instance references exists
             job_instance = await self.client.create_job_instance(
                 CustomerJobInstanceReferenceInput(
                     orchestrator_job_instance_id=job_metadata.get('job_instance_id'),
@@ -179,16 +188,18 @@ class NoteableEngine(Engine):
 
             # 2: Set up the job instance attempt
             # TODO: update the job instance attempt status while running/after completion
-            job_instance_attempt = JobInstanceAttempt(
+            job_instance_attempt_request = JobInstanceAttemptRequest(
                 status=JobInstanceAttemptStatus.CREATED,
                 attempt_number=0,
                 customer_job_instance_reference_id=job_instance.id,
             )
 
         # Create the parameterized_notebook
-        self.file = await self.client.create_parameterized_notebook(
-            original_notebook_id, job_instance_attempt=job_instance_attempt
+        resp = await self.client.create_parameterized_notebook(
+            original_notebook_id, job_instance_attempt=job_instance_attempt_request
         )
+        self.file = resp.parameterized_notebook
+        self.job_instance_attempt = resp.job_instance_attempt
         parameterized_url = f"https://{self.client.config.domain}/f/{self.file.id}"
 
         self.nb.metadata["executed_notebook_url"] = parameterized_url
@@ -208,50 +219,70 @@ class NoteableEngine(Engine):
                 solid_handle=dagster_context.op_handle,
                 event_specific_data=AssetObservationData(asset_obs),
             )
-            dagster_logger.log_dagster_event(
+            ext_logger.log_dagster_event(
                 level="INFO", msg="Parameterized notebook available at", dagster_event=event
             )
         else:
-            dagster_logger.info(f"Parameterized notebook available at {parameterized_url}")
-        # HACK: We need this delay in order to successfully subscribe to the files channel
-        #       of the newly created parameterized notebook.
-        await asyncio.sleep(1)
+            ext_logger.info(f"Parameterized notebook available at {parameterized_url}")
 
-        async with self.setup_kernel(file=self.file, client=self.client, **kwargs):
-            noteable_nb = nbformat.reads(
-                self.file.content
-                if isinstance(self.file.content, str)
-                else json.dumps(self.file.content),
-                as_version=4,
-            )
-            await self.sync_noteable_nb_with_papermill(
-                file=self.file,
-                noteable_nb=noteable_nb,
-                papermill_nb=self.nb,
-                dagster_logger=dagster_logger,
-            )
+        try:
+            async with self.setup_kernel(file=self.file, client=self.client, **kwargs):
+                noteable_nb = nbformat.reads(
+                    self.file.content
+                    if isinstance(self.file.content, str)
+                    else json.dumps(self.file.content),
+                    as_version=4,
+                )
+                await self.sync_noteable_nb_with_papermill(
+                    file=self.file,
+                    noteable_nb=noteable_nb,
+                    papermill_nb=self.nb,
+                    ext_logger=ext_logger,
+                )
 
-            # Sync metadata from papermill to noteable before execution
-            await self.sync_noteable_nb_metadata_with_papermill()
+                # Sync metadata from papermill to noteable before execution
+                await self.sync_noteable_nb_metadata_with_papermill()
 
-            await self.papermill_execute_cells()
+                # We're going to start executing the notebook; Update the job instance attempt status to RUNNING
+                if self.job_instance_attempt:
+                    logger.debug(
+                        f"Updating job instance attempt id {self.job_instance_attempt.id} to status RUNNING"
+                    )
+                    await self.client.update_job_instance(
+                        job_instance_attempt_id=self.job_instance_attempt.id,
+                        job_instance_attempt_update=JobInstanceAttemptUpdate(
+                            status=JobInstanceAttemptStatus.RUNNING
+                        ),
+                    )
 
-            # This is a hack to ensure we have the client in session to send nb metadata
-            # updates over RTU after execution.
-            self.nb_man.notebook_complete()
-            await self.sync_noteable_nb_metadata_with_papermill()
+                await self.papermill_execute_cells()
 
-            # Override the notebook_complete method and set it to a no-op (since we already called it)
-            self.nb_man.notebook_complete = lambda: None
+                # This is a hack to ensure we have the client in session to send nb metadata
+                # updates over RTU after execution.
+                self.nb_man.notebook_complete()
+                await self.sync_noteable_nb_metadata_with_papermill()
 
-            # info_msg = self.wait_for_reply(self.kc.kernel_info())
-            # self.nb.metadata['language_info'] = info_msg['content']['language_info']
+                # Override the notebook_complete method and set it to a no-op (since we already called it)
+                self.nb_man.notebook_complete = lambda: None
+
+                # info_msg = self.wait_for_reply(self.kc.kernel_info())
+                # self.nb.metadata['language_info'] = info_msg['content']['language_info']
+        except:  # noqa
+            logger.exception("Error executing notebook")
+            if self.job_instance_attempt:
+                await self.client.update_job_instance(
+                    job_instance_attempt_id=self.job_instance_attempt.id,
+                    job_instance_attempt_update=JobInstanceAttemptUpdate(
+                        status=JobInstanceAttemptStatus.FAILED
+                    ),
+                )
+            raise
 
         return self.nb
 
     @ensure_client
     async def sync_noteable_nb_with_papermill(
-        self, file: NotebookFile, noteable_nb, papermill_nb, dagster_logger
+        self, file: NotebookFile, noteable_nb, papermill_nb, ext_logger
     ):
         """Used to sync the cells of in-memory notebook representation that papermill manages with the Noteable notebook
 
@@ -271,7 +302,7 @@ class NoteableEngine(Engine):
             after_id = papermill_nb_cell_ids[idx - 1] if idx > 0 else None
             await self.km.client.add_cell(file, cell=papermill_nb.cells[idx], after_id=after_id)
 
-        dagster_logger.info(
+        ext_logger.info(
             "Synced notebook with Noteable, "
             f"added {len(added_cell_ids)} cells and deleted {len(deleted_cell_ids)} cells"
         )
@@ -294,18 +325,14 @@ class NoteableEngine(Engine):
     @asynccontextmanager
     async def setup_kernel(self, cleanup_kc=True, cleanup_kc_on_error=False, **kwargs) -> Generator:
         """Context manager for setting up the kernel to execute a notebook."""
-        dagster_logger = kwargs["logger"]
+        ext_logger = kwargs["logger"]
 
         if self.km is None:
             # Assumes that file and client are being passed in
             self.km = self.create_kernel_manager(**kwargs)
 
-        # Subscribe to the file or we won't see status updates
-        await self.client.subscribe_file(self.km.file, from_version_id=self.file.current_version_id)
-        dagster_logger.info("Subscribed to file")
-
         await self.km.async_start_kernel(**kwargs)
-        dagster_logger.info("Started kernel")
+        ext_logger.info("Started kernel")
 
         try:
             yield
@@ -384,6 +411,7 @@ class NoteableEngine(Engine):
         )
 
         # Execute each cell and update the output in real time.
+        errored = False
         for index, cell in enumerate(self.nb.cells):
             try:
                 self._cell_start(cell, index)
@@ -391,10 +419,35 @@ class NoteableEngine(Engine):
             except CellExecutionError as ex:
                 # TODO: Make sure we raise these
                 self._cell_exception(self.nb.cells[index], index, exception=ex)
+                errored = True
                 break
             finally:
                 self._cell_complete(self.nb.cells[index], cell_index=index)
 
+        # Update the job instance attempt status
+        if self.job_instance_attempt:
+            status = (
+                JobInstanceAttemptStatus.FAILED if errored else JobInstanceAttemptStatus.SUCCEEDED
+            )
+            logger.debug(
+                f"Updating job instance attempt id {self.job_instance_attempt.id} to status {status}"
+            )
+            await self.client.update_job_instance(
+                job_instance_attempt_id=self.job_instance_attempt.id,
+                job_instance_attempt_update=JobInstanceAttemptUpdate(status=status),
+            )
+
+        if not errored:
+            # Delete the kernel session
+            logger.debug("Deleting kernel session for file id %s", self.file.id)
+            try:
+                await self.client.delete_kernel_session(self.file)
+            except httpx.ReadTimeout:
+                logger.warning(
+                    "Timed out while deleting kernel session for file id %s",
+                    self.file.id,
+                )
+
     def _get_timeout(self, cell: Optional[NotebookNode]) -> int:
         """Helper to fetch a timeout as a value or a function to be run against a cell"""
         if self.timeout_func is not None and cell is not None:
